
 { simple procedures to work with the two body problem Kepler elements }

 UNIT UnForKep ;

 INTERFACE

 Uses
   UnConTyp ; { for TypeDim3 or Array [1..3] of Extended }

 Var
   numiters,numitere  : Byte ; { iteration for Kepler equation }

 Function FromAxisToMot ( a  : Extended ) : Extended ;

 Function KeplerEquation ( ElemE , MeanAnom  : Extended ) : Extended ;

 Function KeplerEquaIter ( ElemE , MeanAnom  : Extended ) : Extended ;

 Function TrueAnom ( ElemE , EccAnom  : Extended ) : Extended ;

 Function ToGetTrueAnomaly ( ElemE , MeanAnom  : Extended ) : Extended ;

 Procedure FromCartToKepler ( Pos , Vel  : TypeDim3 ;
                              Var   PosVar , AngVar : TypeDim3 ) ;

 Procedure FromKeplerToCart ( PosVar , AngVar  : TypeDim3 ;
                              Var   Pos , Vel  : TypeDim3 ) ;

 Procedure ForKeplerElem ( ta  : Extended ;  pa , ua  : TypeDim3 ;
                           tb  : Extended ;  Var  pb , ub  : TypeDim3 ) ;

 IMPLEMENTATION

 Uses
   UnForFun ;

 { MeanMotion   : mean motion  n  in revolution per day
   semi-major axis in km  GeoFM in km**3/s**2
   form the formula n**2*a**3=fm }

 Function FromAxisToMot ( a  : Extended ) : Extended ;
 Begin
   FromAxisToMot:=(43200/Pi)*Sqrt(GeoFM/(a*a*a)); { in revolution per day }
 End;

 Function FromMotToAxis ( MeanMotion  : Extended ) : Extended ;
 Var
   mot  : Extended ;
 Begin
   mot:=Pi*MeanMotion/43200; { in radian per second }
   FromMotToAxis:=Exp(1.0/3.0*Ln(GeoFM/Sqr(mot)));
 End;

 { Kepler equation   M=E-e*sin(E)
   to solve with the use Newton iteration method
   from the book
   ...   ᬨ᪮ .
   .,㪠,1965.
     ࠭ 113:
   If  fi(x)=0  then iteration may be possible
   x_(n+1)=f(x_n)  where  f(x)=x-fi(x)/(d/dx(fi(x))) }

 Function KeplerEquation ( ElemE , MeanAnom  : Extended ) : Extended ;
 Var
   iter  : Byte ;
   EccAnom  : Extended ;
   cura,difa  : Extended ;
 Begin
   cura:=MeanAnom+ElemE*Sin(MeanAnom); { in radian }
   EccAnom:=cura; // M=E-e*sin(E) Kepler equation
   difa:=1.0;
   iter:=0;
   While  ( ( difa > 1.0e-15 ) AND ( iter < 241 ) )  Do
     Begin
       EccAnom:=cura-(cura-ElemE*Sin(cura)-MeanAnom)/(1-ElemE*Cos(cura));
       difa:=Abs(cura-EccAnom);
       cura:=EccAnom;
       iter:=iter+1;
     End;
   numitere:=iter+1;
   KeplerEquation:=EccAnom; { eccentric anomaly in radian }
 End;

 { Kepler equation   M=E-e*sin(E)
   to solve with the use simple iteration method }

 Function KeplerEquaIter ( ElemE , MeanAnom  : Extended ) : Extended ;
 Var
   iter  : Byte ;
   EccAnom  : Extended ;
   cura,difa  : Extended ;
 Begin
   cura:=MeanAnom+ElemE*Sin(MeanAnom); { in radian }
   EccAnom:=cura; // M=E-e*sin(E) Kepler equation
   difa:=1.0;
   iter:=0;
   While  ( ( difa > 1.0e-15 ) AND ( iter < 241 ) )  Do
     Begin
       EccAnom:=MeanAnom+ElemE*Sin(cura);
       difa:=Abs(cura-EccAnom);
       cura:=EccAnom;
       iter:=iter+1; { it is necessary only five iterations if e=0.853 }
     End;
   numiters:=iter+1;
   KeplerEquaIter:=EccAnom; { eccentric anomaly in radian }
 End;

 Function TrueAnom ( ElemE , EccAnom  : Extended ) : Extended ;
 Var
   SinE,CosE,SinV,CosV  : Extended ;
 Begin
   SinE:=Sin(EccAnom);
   CosE:=Cos(EccAnom);
   SinV:=Sqrt(1-ElemE*ElemE)*SinE/(1-ElemE*CosE);
   CosV:=(CosE-ElemE)/(1-ElemE*CosE);
   TrueAnom:=EccAnom+ArcTan((SinV*CosE-CosV*SinE)/(CosV*CosE+SinV*SinE));
 End;

 { to get a value of the true anomaly v in degree
   ElemE - eccentricity e , MeanAnom - mean anomaly M in degree }

 Function ToGetTrueAnomaly ( ElemE , MeanAnom  : Extended ) : Extended ;
 Var
   EccA,TrueA,SinE,CosE,SinV,CosV  : Extended ;
 Begin
   EccA:=KeplerEquation(ElemE,GraRad*MeanAnom); { in radian }
   SinE:=Sin(EccA);
   CosE:=Cos(EccA);
   SinV:=Sqrt(1-ElemE*ElemE)*SinE/(1-ElemE*CosE);
   CosV:=(CosE-ElemE)/(1-ElemE*CosE);
   TrueA:=EccA+ArcTan((SinV*CosE-CosV*SinE)/(CosV*CosE+SinV*SinE));
   TrueA:=RadGra*TrueA; { from radian to degree }
   While  TrueA < 0.000  Do  TrueA:=TrueA+360.0;
   While  TrueA > 360.0  Do  TrueA:=TrueA-360.0;
   ToGetTrueAnomaly:=TrueA; { in degree }
 End;

 { to change from
   Descart position and velosity to Kepler osculating elements
   position Pos x,y,z in km 
   velosity Vel vx,vy,vz in km/s
   GeoFM from UnL01010 in km**3/s**2
   is geocentric gravitationl const for the current model
   RadGra from UnL01010 for angle from radian to degree
   PosVar 1,2,3 : a in km, e,  Inclination in degree
   AngVar 1,2,3 : the ascending node in degree
                : the argument of perigei in degree
                : the mean anomaly in degree }

 Procedure FromCartToKepler ( Pos , Vel  : TypeDim3 ;
                              Var   PosVar , AngVar : TypeDim3 ) ;
 Var
   i                  : Byte;
   h,c,l,r,v2,q,
   c1,c2,c3,l1,l2,l3  : Extended ;
   node,argp,
   a,e,p,rincl,
   ArgLat,TrueAnom,
   EccAnom,MeanAnom,
   si,ci,su,cu,sn,cn  : Extended ;
   sp,cp,sv,cv,se,ce  : Extended ;
 Begin
   r:=Sqrt(Sqr(Pos[1])+Sqr(Pos[2])+Sqr(Pos[3])); { in km }
   v2:=Sqr(Vel[1])+Sqr(Vel[2])+Sqr(Vel[3]); { in (km/s)**2 }
   q:=GeoFM/r; { fm in (km**3/s**2)/km from UnL01010 for the Earth }
   h:=v2-2*q; { (km/s)^2 energy integral }
   c1:=Pos[2]*Vel[3]-Pos[3]*Vel[2]; { three momentum integrals }
   c2:=Pos[3]*Vel[1]-Pos[1]*Vel[3];
   c3:=Pos[1]*Vel[2]-Pos[2]*Vel[1];
   c:=Sqrt(c1*c1+c2*c2+c3*c3);
   l1:=-q*Pos[1]+Vel[2]*c3-Vel[3]*c2; { three integrals }
   l2:=-q*Pos[2]+Vel[3]*c1-Vel[1]*c3;
   l3:=-q*Pos[3]+Vel[1]*c2-Vel[2]*c1;
   l:=Sqrt(l1*l1+l2*l2+l3*l3);
   a:=-GeoFM/h; { in km semi-major axis }
   e:=l/GeoFM; { eccentricity }
   p:=c*c/GeoFM;
   ci:=c3/c; si:=Sqrt(Abs(1-ci*ci));
   If  si < 1.0e-11  Then
     Begin
       si:=1.0e-11; rincl:=0; { singularity inclination = 0 }
     End
                     Else  rincl:=DATan2(si,ci); { inclination in radian }
   sn:=c1/(c*si); cn:=-c2/(c*si);
   node:=DATan2(sn,cn); { node in radian }
   sp:=l3/(l*si); cp:=(l1/l)*cn+(l2/l)*sn;
   argp:=DATan2(sp,cp); { argument of perigei in radian }
   q:=1/r; su:=q*Pos[3]/si; cu:=q*Pos[1]*cn+q*Pos[2]*sn;
   sv:=su*cp-cu*sp; cv:=cu*cp+su*sp; { argument of latitude u = v + omega }
   TrueAnom:=DATan2(sv,cv); { true anomaly in radian }
   q:=1/(1+e*cv); se:=Sqrt(1-e*e)*sv*q; ce:=(cv+e)*q;
   EccAnom:=TrueAnom+ArcTan((se*cv-ce*sv)/(ce*cv+se*sv));
   MeanAnom:=EccAnom-e*se; { M = E - e * Sin(E) in radian }
   PosVar[1]:=a; { in km }
   PosVar[2]:=e;
   PosVar[3]:=RadGra*rincl; { in degree }
   AngVar[1]:=RadGra*node; { in degree }
   AngVar[2]:=RadGra*argp; { in degree }
   AngVar[3]:=RadGra*MeanAnom; { in degree }
   For i:=1 To 3 Do
     Begin
       While  AngVar[i] < -180  Do  AngVar[i]:=AngVar[i]+360;
       While  AngVar[i] > +180  Do  AngVar[i]:=AngVar[i]-360;
     End;
 End;

 { to change from
   Kepler osculating elements to Descart position and velosity
   GeoFM from UnL01010 in km**3/s**2
   is geocentric gravitationl const for the current model
   RadGra from UnL01010 for angle from radian to degree
   PosVar 1,2,3 : a in km, e,  Inclination in degree
   AngVar 1,2,3 : the ascending node in degree
                : the argument of perigei in degree
                : the mean anomaly un degree
   position Pos 1,2,3  : x,y,z in km 
   velosity Vel 1,2,3  : vx,vy,vz in km/s }

 Procedure FromKeplerToCart ( PosVar , AngVar  : TypeDim3 ;
                              Var   pos , vel  : TypeDim3 ) ;
 Var
   q,vr,vn  : Extended ;
   EccAnom  : Extended ;
   MeanAnom  : Extended ;
   TrueAnom  : Extended ;
   ArgLatitude  : Extended ;
   AscendingNode  : Extended ;
   r,xr,yr,zr,xv,yv,zv  : Extended ;
   ElemA,ElemE,CosInc,SinInc  : Extended ;
   CosE,SinE,CosV,SinV,CosU,SinU,CosNode,SinNode  : Extended ;
 Begin
   ElemA:=PosVar[1];
   ElemE:=PosVar[2];
   CosInc:=Cos(GraRad*PosVar[3]);
   SinInc:=Sin(GraRad*PosVar[3]);
   MeanAnom:=GraRad*AngVar[3]; { to radian }
   EccAnom:=KeplerEquation(ElemE,MeanAnom);
   SinE:=Sin(EccAnom);
   CosE:=Cos(EccAnom);
   SinV:=Sqrt(1-ElemE*ElemE)*SinE/(1-ElemE*CosE);
   CosV:=(CosE-ElemE)/(1-ElemE*CosE);
   TrueAnom:=EccAnom+ArcTan((SinV*CosE-CosV*SinE)/(CosV*CosE+SinV*SinE));
   ArgLatitude:=TrueAnom+GraRad*AngVar[2]; { in radian }
   SinU:=Sin(ArgLatitude);
   CosU:=Cos(ArgLatitude);
   AscendingNode:=GraRad*AngVar[1];
   SinNode:=Sin(AscendingNode);
   CosNode:=Cos(AscendingNode);
   r:=ElemA*(1-ElemE*CosE); { in km }
   xr:=+CosU*CosNode-SinU*SinNode*CosInc;
   yr:=+CosU*SinNode+SinU*CosNode*CosInc;
   zr:=+SinU*SinInc;
   xv:=-SinU*CosNode-CosU*SinNode*CosInc;
   yv:=-SinU*SinNode+CosU*CosNode*CosInc;
   zv:=+CosU*SinInc;
   q:=Sqrt(GeoFM/(ElemA*(1-Sqr(ElemE))));
   vr:=q*ElemE*SinV;
   vn:=q*(1+ElemE*CosV);
   pos[1]:=r*xr;
   pos[2]:=r*yr;
   pos[3]:=r*zr;
   vel[1]:=vr*xr+vn*xv;
   vel[2]:=vr*yr+vn*yv;
   vel[3]:=vr*zr+vn*zv;
 End;

 Procedure ForKeplerElem ( ta  : Extended ;  pa , ua  : TypeDim3 ;
                           tb  : Extended ;  Var  pb , ub  : TypeDim3 ) ;
 Var
   n  : Extended ;
 Begin
   n:=FromAxisToMot(pa[1])*(360/86400); { mean motion in degree per second }
   pb:=pa; { the same elements }
   ub:=ua; { the same elements }
   ub[3]:=ua[3]+n*(tb-ta); { mean anomaly M in degree to new moment }
 End;

 END.